#include <os.h>
#include <nspireio2.h>

#include "math.h"

#define fillScr(color) memset(SCREEN_BASE_ADDRESS, color + (color << 4), SCREEN_BYTES_SIZE);

#define MAP_SIZE_X 64
#define MAP_SIZE_Y 64

#define TEX_SIZE 16

void mode7BuildLut(vec2_t* lutPtr, int screenHeight, int screenWidth, float planeH, float planeV, float height) {
	int screenRow;
	for (screenRow = 0; screenRow < screenHeight; screenRow++) {
		float cameraV = (((float)screenRow / (float)screenHeight) * 2.0 - 1.0);
		
		float rayX = planeH;
		float rayY = 1.0;
		float rayZ = planeV * cameraV;
		
		float intersectY = (height / rayZ) * rayY;
		float intersectX = -(height / rayZ) * rayX;
		
		lutPtr[screenRow].x = (int32_t)(intersectX * 65536);
		lutPtr[screenRow].y = (int32_t)(intersectY * 65536); 
	}
}

void mode7Render(char* scrbuf, vec2_t* lutPtr, int screenHeight, int screenWidth, char* tilemap, int mapSizeX, int mapSizeY, char** texIdx, int texSize, vec2_t pos, int32_t heading) {
	int screenX, screenY;
	for (screenY = screenHeight / 2; screenY < screenHeight; screenY++) {
		vec2_t map = lutPtr[screenY];
		vec2_t mapXStep = {-map.x / (screenWidth / 2), 0};
		
		int32_t sinHeading = sin(heading);
		int32_t cosHeading = cos(heading);
		
		map = rotateVec2(map, sinHeading, cosHeading);
		mapXStep = rotateVec2(mapXStep, sinHeading, cosHeading);
		
		map.x += pos.x;
		map.y += pos.y;
		
		for (screenX = 0; screenX < SCREEN_WIDTH; screenX++) {
			int color = 0;
			//map.x = map.x && numToFix(mapSizeX) - 1;
			//map.y = map.y && numToFix(mapSizeY) - 1;
			if (map.x >= 0 && map.x < numToFix(mapSizeX) && map.y >= 0 && map.y < numToFix(mapSizeY)) {
			//if (map.x == map.x & (numToFix(MAP_SIZE_X) - 1) && map.y == map.y & (numToFix(MAP_SIZE_Y) - 1)) {
				int32_t mapIdx = (map.y >> 16) * 64 + (map.x >> 16);
				color = tilemap[mapIdx];
				if (color > 0xF) {
				    color = texIdx[color - 16][((map.y & 0xFFFF) >> 12) * texSize + ((map.x & 0xFFFF) >> 12)];
				}
			}
			//if (map.x >= 0 && map.x < numToFix(MAP_SIZE_X) && map.y >= 0 && map.y < numToFix(MAP_SIZE_Y)) color >>= 2;
						
			map.x += mapXStep.x;
			map.y += mapXStep.y;
			
			setPixel(screenX, screenY, color);
		}
	}
}

int main(void) {
    int32_t i, j, k; //for loop counters
	
	//nio_console csl;
	//lcd_ingray();
    //clrscr();
    //53 columns, 29 rows. 0px offset for x/y.
    //Background color 0 (black), foreground color 15 (white)
    //nio_InitConsole(&csl, 53, 29, 0, 0, 0, 15);
	fillScr(0x0);
    putStr(0, 0, "mode 7 test by fb39ca4", 0x0, 0xF);

    char* tilemap = (char*) malloc(MAP_SIZE_X * MAP_SIZE_Y);
	srand(42);
	for (i = 0; i < MAP_SIZE_X * MAP_SIZE_Y; i++) tilemap[i] = rand() % 20;
	
	char* xorTex = (char*) malloc(16 * 16);
	for (i = 0; i < 16; i++) for (j = 0; j < 16; j++) xorTex[j * 16 + i] = i ^ j;
	
	char gradient1Tex[16][16];
	for (i = 0; i < 16; i++) for (j = 0; j < 16; j++) gradient1Tex[i][j] = i;
	
	char gradient2Tex[16][16];
	for (k = 0; k < 8; k++) {
		for (i = k; i < 16 - k; i++) for (j = k; j < 16 - k; j++) gradient2Tex[i][j] = k << 1;
	}
	
	char concentricTex[16][16];
	for (k = 0; k < 8; k++) {
		for (i = k; i < 16 - k; i++) for (j = k; j < 16 - k; j++) concentricTex[i][j] = (k & 1) * 8 + 4;
	}
	
	char* textureIdx[4] = {xorTex, gradient1Tex, gradient2Tex, concentricTex};
	
	vec2_t* mode7lut = (vec2_t*) malloc(sizeof(vec2_t) * SCREEN_HEIGHT);
	
	float planeH = 1.0;
	float planeV = 0.75;
	float height = 1;
	
	mode7BuildLut(mode7lut, SCREEN_HEIGHT, SCREEN_WIDTH, planeH, planeV, height);
		
	uint16_t time = 0;
	uint16_t oldTime = 0;
	
	vec2_t pos = {numToFix(4), numToFix(4)};
	
	int32_t heading = 0;
		
	char fpsString[8];
	
	while(!isKeyPressed(KEY_NSPIRE_ESC)) {
		oldTime = time;
		time = *(volatile uint16_t*) 0x900D000C;
		uint32_t fps = 32768 / ((time - oldTime) & 0xFFFF);
		sprintf(fpsString, "%4u FPS", fps);
		putStr(270, 0, fpsString, 0x0, 0xF);
		
		int32_t sinA = sin(heading);
		int32_t cosA = cos(heading);
		
		vec2_t moveDist = rotateVec2((vec2_t){numToFix(0), floatToFix(0.25)}, sinA, cosA);
		
		if (isKeyPressed(KEY_NSPIRE_W)) pos = addVec2(pos, moveDist);
		if (isKeyPressed(KEY_NSPIRE_S)) pos = subVec2(pos, moveDist);
		if (isKeyPressed(KEY_NSPIRE_D)) pos = subVec2(pos, rotateVec2(moveDist, numToFix(1), 0));
		if (isKeyPressed(KEY_NSPIRE_A)) pos = addVec2(pos, rotateVec2(moveDist, numToFix(1), 0));
		if (isKeyPressed(KEY_NSPIRE_E)) heading += 0x400;
		if (isKeyPressed(KEY_NSPIRE_Q)) heading -= 0x400;
		
		
		mode7Render(SCREEN_BASE_ADDRESS, mode7lut, SCREEN_HEIGHT, SCREEN_WIDTH, 
					tilemap, MAP_SIZE_X, MAP_SIZE_Y,
					textureIdx, TEX_SIZE,
					pos, heading);
		
	}
	
	//nio_CleanUp(&csl);
	free(mode7lut);
	free(tilemap);
	return 0;
}